//******************************************************************************
// Copyright (c) 2015 Dave Hein
// See end of file for terms of use.
//******************************************************************************
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "gencode.h"
#include "cstrsubs.h"

FILE *outfile;
int reg_indx;
char callstack[100];
int cs_index = 0;
int nopflag;
int picflag;
int globalflag;

static char size2char[] = {'v', 'b', 'w', 'x', 'l'};

#define USE_CALLD

// Write the Spasm header
void EmitHeader(void)
{
    Emit("dat\n");
    Emit("        orgh    0\n");
    Emit("        org     0\n");
    Emit("        cmp     ptra, #0  wz\n");
    Emit(" if_nz  mov     sp, ptra\n");
    Emit(" if_z   mov     sp, ##$7ff80\n");
    Emit("        or      outb, ##$40000000\n");
    Emit("        or      dirb, ##$40000000\n");
    if (picflag)
    {
        Emit("        mov     reg0, pictable\n");
        Emit("        add     reg0, ptrb\n");
        Emit("        rdlong  reg1, reg0 wz\n");
        Emit(" if_z   jmp     #adjustaddr\n");
        Emit("        add     reg0, #4\n");
        Emit(".loop   rdlong  pictable, reg0\n");
        Emit("        add     pictable, ptrb\n");
        Emit("        add     .loop, ##$200\n");
        Emit("        add     .loop+1, ##$200\n");
        Emit("        add     reg0, #4\n");
        Emit("        djnz    reg1, #.loop\n");
        Emit("adjustaddr\n");
        Emit("        mov     reg0, adjtable\n");
        Emit("        add     reg0, ptrb\n");
        Emit("        rdlong  reg1, reg0 wz\n");
        Emit("        add     reg0, #4\n");
        Emit(" if_z   jmp     #callmain\n");
        Emit(".loop   rdlong  reg2, reg0\n");
        Emit("        add     reg2, ptrb\n");
        Emit("        rdlong  reg3, reg2 wz\n");
        Emit("        add     reg3, ptrb\n");
        Emit("        wrlong  reg3, reg2\n");
        Emit("        add     reg0, #4\n");
        Emit("        djnz    reg1, #.loop\n");
        Emit("callmain\n");
        Emit("        mov     reg0, ##main\n");
        Emit("        add     reg0, ptrb\n");
        Emit("        calld   adra, reg0\n");
    }
    else
    {
        Emit("        calld   adra, #main\n");
    }
    Emit("        cogid   reg0\n");
    Emit("        mov     reg1, reg0\n");
    Emit("        shl     reg1, #2\n");
    Emit("        add     reg1, ##$7ff80\n");
    Emit("        wrlong  #0, reg1\n");
    Emit("        cogstop reg0\n");
    Emit("sp      long    0\n");
    Emit("reg0    long    0\n");
    Emit("reg1    long    0\n");
    Emit("reg2    long    0\n");
    Emit("reg3    long    0\n");
    Emit("reg4    long    0\n");
    Emit("reg5    long    0\n");
    Emit("reg6    long    0\n");
    Emit("reg7    long    0\n");
    Emit("reg8    long    0\n");
    Emit("reg9    long    0\n");
    Emit("reg10   long    0\n");
    Emit("reg11   long    0\n");
    Emit("reg12   long    0\n");
    Emit("reg13   long    0\n");
    Emit("reg14   long    0\n");
    Emit("reg15   long    0\n");
    Emit("\n");
    Emit("\n");
    Emit("_divide_\n");
    Emit("        mov     temp1, #0 wz\n");
    Emit("        cmps    parm1, #0 wc\n");
    Emit("        abs     parm1, parm1\n");
    Emit("  if_c  mov     temp1, #1 wz\n");
    Emit("        cmps    parm2, #0 wc\n");
    Emit("        abs     parm2, parm2\n");
    Emit("  if_c  xor     temp1, #1 wz\n");
    Emit("        abs     parm2, parm2\n");
    Emit("        qdiv    parm1, parm2\n");
    Emit("        getqx   parm1\n");
    Emit("  if_nz neg     parm1, parm1\n");
    Emit("        ret\n");
    Emit("\n");
    Emit("_modulus_\n");
    Emit("        mov     temp1, #0 wz\n");
    Emit("        cmps    parm1, #0 wc\n");
    Emit("        abs     parm1, parm1\n");
    Emit("  if_c  mov     temp1, #1 wz\n");
    Emit("        abs     parm2, parm2\n");
    Emit("        qdiv    parm1, parm2\n");
    Emit("        getqy   parm1\n");
    Emit("  if_nz neg     parm1, parm1\n");
    Emit("        ret\n");
    Emit("\n");
    Emit("temp1   long    0\n");
    Emit("temp2   long    0\n");
    Emit("parm1   long    0\n");
    Emit("parm2   long    0\n");
    Emit("parm3   long    0\n");
    if (picflag)
    {
        Emit("adjtable\n");
        Emit("        long    0\n");
        Emit("pictable\n");
        Emit("        long    _hub_pic_table\n");
    }
    Emit("\n");
    Emit("        orgh    $400\n");
    Emit("\n");
}

static char *GetTypeStrFromChar(int val)
{
    char *typestr = "????";
    if (val == 'b') typestr = "byte";
    else if (val == 'w') typestr = "word";
    else if (val == 'l') typestr = "long";
    return typestr;
}

void Emit(char *fmtstr)
{
    fprintf(outfile, fmtstr);
    SetPrintNL();
}

void Emit1(char *fmtstr, char *parm1)
{
    fprintf(outfile, fmtstr, parm1);
    SetPrintNL();
}

void Emit1i(char *fmtstr, int parm1)
{
    fprintf(outfile, fmtstr, parm1);
    SetPrintNL();
}

void Emit2(char *fmtstr, char *parm1, char *parm2)
{
    fprintf(outfile, fmtstr, parm1, parm2);
    SetPrintNL();
}

void Emit3(char *fmtstr, char *parm1, char *parm2, char *parm3)
{
    fprintf(outfile, fmtstr, parm1, parm2, parm3);
    SetPrintNL();
}

void Emit4(char *fmtstr, char *parm1, char *parm2, char *parm3, char *parm4)
{
    fprintf(outfile, fmtstr, parm1, parm2, parm3, parm4);
    SetPrintNL();
}

void Emit5(char *fmtstr, char *parm1, char *parm2, char *parm3, char *parm4, char *parm5)
{
    fprintf(outfile, fmtstr, parm1, parm2, parm3, parm4, parm5);
    SetPrintNL();
}

void EmitNops(int num)
{
    int i;

    if (!num) num = nopflag;

    for (i = 0; i < num; i++)
        Emit("        nop\n");
}

void EmitJump(int num)
{
    Emit1i("        jmp     #label%04d\n", num);
    EmitNops(0);
}

void EmitJumpOnZero(int num)
{
    Emit1i("        cmp     reg%d, #0  wz\n", --reg_indx);
    Emit1i(" if_z   jmp     #label%04d\n", num);
    EmitNops(0);
}

void EmitJumpOnNonZero(int num)
{
    Emit1i("        cmp     reg%d, #0  wz\n", --reg_indx);
    Emit1i(" if_nz  jmp     #label%04d\n", num);
    EmitNops(0);
}

void EmitLabel(int num)
{
    Emit1i("label%04d\n", num);
}

void EmitLoadReg(char *varname)
{
    Emit2("        mov     reg%d, %s\n", (char *)(reg_indx++), varname);
}

void EmitLoadVarAddress(char *varname, int value, int local)
{
    if (local == 1)
    {
        printf("%d: ERROR: Can't get address of register variable %s\n", GetLineNumber(), varname);
        PrintLine();
        Emit1("ERROR: Can't get address of register variable %s\n", varname);
        Emit1i("        mov     reg%d, #0\n", reg_indx++);
    }
    else if (local == 2)
    {
        if (value*4 < 512)
            Emit2("        mov     reg%d, #%d\n", (char *)reg_indx, (char *)(value*4));
        else
            Emit2("        mov     reg%d, ##%d\n", (char *)reg_indx, (char *)(value*4));
        Emit1("        add     reg%d, sp\n", (char *)(reg_indx++));
    }
    else
    {
        if (picflag)
            Emit2("        mov     reg%d, pictable+%d\n", (char *)(reg_indx++), (char *)value);
        else
            Emit2("        mov     reg%d, ##%s\n", (char *)(reg_indx++), varname);
    }
}

void EmitLoadRegIndexed(void)
{
    Emit1i("        alts    reg%d, #reg0\n", reg_indx-1);
    Emit1i("        mov     reg%d, 0-0\n", reg_indx-1);
}

void EmitLoadNumberStr(char *numstr)
{
    int val;
    if (strlen(numstr) >= 2 && numstr[1] == 'x')
    {
        sscanf(numstr+2, "%x", &val);
        if (val < 512 && val >= 0)
            Emit2("        mov     reg%d, #$%s\n", (char *)(reg_indx++), numstr+2);
        else
            Emit2("        mov     reg%d, ##$%s\n", (char *)(reg_indx++), numstr+2);
    }
    else
    {
        sscanf(numstr, "%d", &val);
        if (val < 512 && val >= 0)
            Emit2("        mov     reg%d, #%s\n", (char *)(reg_indx++), numstr);
        else
            Emit2("        mov     reg%d, ##%s\n", (char *)(reg_indx++), numstr);
    }
}

void EmitLoadNumber(int val)
{
    if (val < 512 && val >= 0)
        Emit2("        mov     reg%d, #%d\n", (char *)(reg_indx++), (char *)val);
    else
        Emit2("        mov     reg%d, ##%d\n", (char *)(reg_indx++), (char *)val);
}

void EmitNegate(void)
{
    Emit2("        neg     reg%d, reg%d\n", (char *)(reg_indx-1), (char *)(reg_indx-1));
}

void EmitLogicalNot(void)
{
    Emit1("        cmp     reg%d, #0  wz\n", (char *)(reg_indx-1));
    Emit1(" if_z   mov     reg%d, #1\n", (char *)(reg_indx-1));
    Emit1(" if_nz  mov     reg%d, #0\n", (char *)(reg_indx-1));
}

void EmitComplement(void)
{
    Emit1i("        xor     reg%d, ##$ffffffff\n", reg_indx-1);
}

void EmitLoadVar(char *varname, int signflag, int size, int pointer, int value)
{
    int parmchar = 'l';

    if (!pointer)
        parmchar = size2char[size];
   
    if (picflag)
        Emit3("        rd%s  reg%d, pictable+%d\n", GetTypeStrFromChar(parmchar), (char *)(reg_indx++), (char *)value);
    else 
        Emit3("        rd%s  reg%d, ##%s\n", GetTypeStrFromChar(parmchar), (char *)(reg_indx++), varname);
    if (parmchar != 'l' && signflag)
            EmitSignExtend(parmchar, reg_indx-1);
}

void EmitLoadVariable(char *varname, int signflag, int size, int pointer, int array, int local, int value)
{
    if (array)
        EmitLoadVarAddress(varname, value, local);
    else if (local == 1)
        EmitLoadParm(value);
    else if (local == 2)
        EmitLoadLocal(value);
    else
        EmitLoadVar(varname, signflag, size, pointer, value);
}

void EmitLoadParm(int num)
{
    Emit2("        mov     reg%d, reg%d\n", (char *)(reg_indx++), (char *)num);
}

void EmitLoadLocal(int num)
{
    if (num*4 < 512)
        Emit2("        mov     reg%d, #%d\n", (char *)reg_indx, (char *)(num * 4));
    else
        Emit2("        mov     reg%d, ##%d\n", (char *)reg_indx, (char *)(num * 4));
    Emit1i("        add     reg%d, sp\n", reg_indx);
    Emit2("        rdlong  reg%d, reg%d\n", (char *)reg_indx, (char *)reg_indx);
    reg_indx++;
}

#if 0
void EmitLoadAddress(char *varname)
{
    Emit2("        mov     reg%d, ##%s\n", (char *)(reg_indx++), varname);
}
#endif

void EmitLoadAbsolute(int parmchar)
{
    Emit3("        rd%s  reg%d, reg%d\n", GetTypeStrFromChar(parmchar), (char *)(reg_indx-1), (char *)(reg_indx-1));
}

void EmitAddAbsoluteIndexed(int parmchar)
{
    int shiftval = 0;
    if (parmchar == 'l') shiftval = 2;
    else if (parmchar == 'w') shiftval = 1;
    if (shiftval)
        Emit2("        shl     reg%d, #%d\n", (char *)(reg_indx-1), (char *)shiftval);
    reg_indx--;
    Emit2("        add     reg%d, reg%d\n", (char *)(reg_indx-1), (char *)reg_indx);
}

void EmitLoadAbsoluteIndexed(int parmchar)
{
    int shiftval = 0;
    if (parmchar == 'l') shiftval = 2;
    else if (parmchar == 'w') shiftval = 1;
    if (shiftval)
        Emit2("        shl     reg%d, #%d\n", (char *)(reg_indx-1), (char *)shiftval);
    reg_indx--;
    Emit2("        add     reg%d, reg%d\n", (char *)(reg_indx-1), (char *)reg_indx);
    Emit3("        rd%s  reg%d, reg%d\n", GetTypeStrFromChar(parmchar), (char *)(reg_indx-1), (char *)(reg_indx-1));
}

void EmitSignExtend(int parmchar, int reg_indx1)
{
    int shlen = (parmchar == 'b') ? 24 : 16;
    Emit2("        shl     reg%d, #%d\n", (char *)reg_indx1, (char *)shlen);
    Emit2("        sar     reg%d, #%d\n", (char *)reg_indx1, (char *)shlen);
}

void CallFunc(char *funcname)
{
#ifdef USE_CALLD
    Emit1("        calld   adra, #%s\n", funcname);
#else
    Emit1("        call    #%s\n", funcname);
#endif
    EmitNops(0);
}

void EmitOpcode(int opdex, int signflag)
{
    reg_indx--;
    switch (opdex)
    {
        case OP_MULTIPY:
            Emit2("        qmul    reg%d, reg%d\n", (char *)(reg_indx-1), (char *)reg_indx);
            Emit1("        getqx   reg%d\n", (char *)(reg_indx-1));
            break;
        case OP_DIVIDE:
            // TODO: Handle unsigned divide
            Emit1i("        mov     parm1, reg%d\n", reg_indx-1);
            Emit1i("        mov     parm2, reg%d\n", reg_indx);
            Emit("        call    #_divide_\n");
            Emit1i("        mov     reg%d, parm1\n", reg_indx-1);
            break;
        case OP_MOD:
            // TODO: Handle unsigned modulus
            Emit1i("        mov     parm1, reg%d\n", reg_indx-1);
            Emit1i("        mov     parm2, reg%d\n", reg_indx);
            Emit("        call    #_modulus_\n");
            Emit1i("        mov     reg%d, parm1\n", reg_indx-1);
            break;
        case OP_PLUS:
            Emit2("        add     reg%d, reg%d\n", (char *)(reg_indx-1), (char *)reg_indx);
            break;
        case OP_MINUS:
            Emit2("        sub     reg%d, reg%d\n", (char *)(reg_indx-1), (char *)reg_indx);
            break;
        case OP_SHIFTL:
            Emit2("        shl     reg%d, reg%d\n", (char *)(reg_indx-1), (char *)reg_indx);
            break;
        case OP_SHIFTR:
            if (signflag)
                Emit2("        sar     reg%d, reg%d\n", (char *)(reg_indx-1), (char *)reg_indx);
            else
                Emit2("        shr     reg%d, reg%d\n", (char *)(reg_indx-1), (char *)reg_indx);
            break;
        case OP_CMPLT:
            if (signflag)
                Emit2("        cmps    reg%d, reg%d  wc\n", (char *)(reg_indx-1), (char *)reg_indx);
            else
                Emit2("        cmp     reg%d, reg%d  wc\n", (char *)(reg_indx-1), (char *)reg_indx);
            Emit1(" if_c   mov     reg%d, #1\n", (char *)(reg_indx-1));
            Emit1(" if_nc  mov     reg%d, #0\n", (char *)(reg_indx-1));
            break;
        case OP_CMPLE:
            if (signflag)
                Emit2("        cmps    reg%d, reg%d wc\n", (char *)reg_indx, (char *)(reg_indx-1));
            else
                Emit2("        cmp     reg%d, reg%d wc\n", (char *)reg_indx, (char *)(reg_indx-1));
            Emit1(" if_nc  mov     reg%d, #1\n", (char *)(reg_indx-1));
            Emit1(" if_c   mov     reg%d, #0\n", (char *)(reg_indx-1));
            break;
        case OP_CMPGT:
            if (signflag)
                Emit2("        cmps    reg%d, reg%d wc\n", (char *)reg_indx, (char *)(reg_indx-1));
            else
                Emit2("        cmp     reg%d, reg%d wc\n", (char *)reg_indx, (char *)(reg_indx-1));
            Emit1(" if_c   mov     reg%d, #1\n", (char *)(reg_indx-1));
            Emit1(" if_nc  mov     reg%d, #0\n", (char *)(reg_indx-1));
            break;
        case OP_CMPGE:
            if (signflag)
                Emit2("        cmps    reg%d, reg%d  wc\n", (char *)(reg_indx-1), (char *)reg_indx);
            else
                Emit2("        cmp     reg%d, reg%d  wc\n", (char *)(reg_indx-1), (char *)reg_indx);
            Emit1(" if_nc  mov     reg%d, #1\n", (char *)(reg_indx-1));
            Emit1(" if_c   mov     reg%d, #0\n", (char *)(reg_indx-1));
            break;
        case OP_CMPEQ:
            Emit2("        cmp     reg%d, reg%d  wz\n", (char *)(reg_indx-1), (char *)reg_indx);
            Emit1(" if_z   mov     reg%d, #1\n", (char *)(reg_indx-1));
            Emit1(" if_nz  mov     reg%d, #0\n", (char *)(reg_indx-1));
            break;
        case OP_CMPNE:
            Emit2("        sub     reg%d, reg%d  wz\n", (char *)(reg_indx-1), (char *)reg_indx);
            Emit1(" if_nz  mov     reg%d, #1\n", (char *)(reg_indx-1));
            break;
        case OP_AND:
            Emit2("        and     reg%d, reg%d\n", (char *)(reg_indx-1), (char *)reg_indx);
            break;
        case OP_XOR:
            Emit2("        xor     reg%d, reg%d\n", (char *)(reg_indx-1), (char *)reg_indx);
            break;
        case OP_OR:
            Emit2("        or      reg%d, reg%d\n", (char *)(reg_indx-1), (char *)reg_indx);
            break;
        case OP_ANDL:
            Emit1("        cmp     reg%d, #0  wz\n", (char *)(reg_indx-1));
            Emit1(" if_nz  cmp     reg%d, #0  wz\n", (char *)reg_indx);
            Emit1(" if_nz  mov     reg%d, #1\n", (char *)(reg_indx-1));
            Emit1(" if_z   mov     reg%d, #0\n", (char *)(reg_indx-1));
            break;
        case OP_ORL:
            Emit2("        or      reg%d, reg%d  wz\n", (char *)(reg_indx-1), (char *)reg_indx);
            Emit1(" if_nz  mov     reg%d, #1\n", (char *)(reg_indx-1));
            break;
    }
}

char *SetupVarAccess(char *varname, int absolute, int indexed, int parmchar, int local, int value, char *addrstr)
{
    char *typestr = GetTypeStrFromChar(parmchar);
    int shiftval = 0;
    if (parmchar == 'l') shiftval = 2;
    else if (parmchar == 'w') shiftval = 1;

    strcpy(addrstr, "????");

    if (parmchar == 'r')
    {
        strcpy(addrstr, varname);
    }
    else if (local == 1 && !absolute)
    {
        sprintf(addrstr, "reg%d", value);
    }
    else if (local == 2 && !absolute)
    {
        sprintf(addrstr, "reg%d", reg_indx);
        if (value*4 < 512)
            Emit2("        mov     %s, #%d\n", addrstr, (char *)(value*4));
        else
            Emit2("        mov     %s, ##%d\n", addrstr, (char *)(value*4));
        Emit1("        add     %s, sp\n", addrstr);
    }
    else if (absolute)
    {
        if (indexed)
        {
            reg_indx -= 3;
            if (shiftval) Emit2("        shl     reg%d, #%d\n", (char *)(reg_indx+2), (char *)shiftval);
            Emit2("        add     reg%d, reg%d\n", (char *)(reg_indx+1), (char *)(reg_indx+2));
            sprintf(addrstr, "reg%d", reg_indx+1);
            reg_indx++;
        }
        else
        {
            reg_indx -= 2;
            sprintf(addrstr, "reg%d", reg_indx+1);
            reg_indx++;
        }
    }
    else
    {
        if (indexed)
        {
            reg_indx -= 2;
            if (shiftval) Emit2("        shl     reg%d, #%d\n", (char *)(reg_indx+1), (char *)shiftval);
            if (picflag)
                Emit2("        add     reg%d, pictable+%d\n", (char *)(reg_indx+1), (char *)value);
            else
                Emit2("        add     reg%d, ##%s\n", (char *)(reg_indx+1), varname);
            sprintf(addrstr, "reg%d", reg_indx+1);
            reg_indx++;
        }
        else
        {
            if (picflag)
                sprintf(addrstr, "pictable+%d", value);
            else
                sprintf(addrstr, "##%s", varname);
        }
    }
    return typestr;
}

void EmitReadAccess(char *addrstr, char *typestr, int parmchar, int local, int signflag, int reg_indx1)
{
    if (parmchar == 'r' || local == 1)
        Emit2("        mov     reg%d, %s\n", (char *)reg_indx1, addrstr);
    else
    {
        Emit3("        rd%s  reg%d, %s\n", typestr, (char *)reg_indx1, addrstr);
        if (parmchar != 'l' && signflag) EmitSignExtend(parmchar, reg_indx1);
    }
}

void EmitWriteAccess(char *addrstr, char *typestr, int parmchar, int local, int reg_indx1, int absolute)
{
//fprintf(outfile, "EmitWriteAccess: addrstr = %s, typestr = %s, parmchar = %c, local = %d\n", addrstr, typestr, parmchar, local);
    if (parmchar == 'r' || (local == 1 && !absolute))
        Emit2("        mov     %s, reg%d\n", addrstr, (char *)reg_indx1);
    else
    {
        Emit3("        wr%s  reg%d, %s\n", typestr, (char *)reg_indx1, addrstr);
    }
}

void EmitIncrement(char *varname, int post, int incrsize, int plus, int retflag, int size, int pointer, int signflag, int absolute, int indexed, int parmchar, int local, int value)
{
    int reg_indx0 = reg_indx++;
    int reg_indx1 = reg_indx++;
    char addrstr[40];
    char *typestr = SetupVarAccess(varname, absolute, indexed, parmchar, local, value, addrstr);

    EmitReadAccess(addrstr, typestr, parmchar, local, signflag, reg_indx0);

    if (retflag && post)
        Emit2("        mov     reg%d, reg%d\n", (char *)reg_indx1, (char *)reg_indx0);
    else
        reg_indx1 = reg_indx0;

    if (plus)
        Emit2("        add     reg%d, #%d\n", (char *)reg_indx1, (char *)incrsize);
    else
        Emit2("        sub     reg%d, #%d\n", (char *)reg_indx1, (char *)incrsize);

    EmitWriteAccess(addrstr, typestr, parmchar, local, reg_indx1, absolute);

    reg_indx--;
    if (!retflag) reg_indx--;
}

void EmitVariableAssignment(char *varname, int index, int retflag, int absolute, int indexed, int parmchar, int signflag, int local, int value)
{
    int reg_indx1;
    char addrstr[40];
    char *typestr = SetupVarAccess(varname, absolute, indexed, parmchar, local, value, addrstr);

    reg_indx--;
    reg_indx1 = reg_indx+2;

//fprintf(outfile, "EmitVariableAssignment: reg_indx = %d, varname = %s, absolute = %d, indexed = %d, parmchar = %d, local = %d, value = %d, addrstr = %s\n",
//reg_indx, varname, absolute, indexed, parmchar, local, value, addrstr);

    if (index)
    {
        EmitReadAccess(addrstr, typestr, parmchar, local, signflag, reg_indx1);
        switch (index)
        {
            case AS_NULL:
                break;
            case AS_PLUS:
                Emit2("        add     reg%d, reg%d\n", (char *)reg_indx, (char *)reg_indx1);
                reg_indx1 = reg_indx;
                break;
            case AS_MINUS:
                Emit2("        sub     reg%d, reg%d\n", (char *)reg_indx1, (char *)reg_indx);
                break;
            case AS_MULT:
#if 0
                Emit1i("        mov     parm1, reg%d\n", reg_indx);
                Emit1i("        mov     parm2, reg%d\n", reg_indx1);
                Emit("        call    #multiply\n");
                EmitNops(2);
                Emit1i("        mov     reg%d, parm1\n", reg_indx);
#else
                Emit2("        qmul    reg%d, reg%d\n", (char *)reg_indx, (char *)reg_indx1);
                Emit1("        getqx   reg%d\n", (char *)reg_indx);
#endif
                reg_indx1 = reg_indx;
                break;
            case AS_DIVIDE:
                // TODO: Handle unsigned divide
                Emit1i("        mov     parm1, reg%d\n", reg_indx1);
                Emit1i("        mov     parm2, reg%d\n", reg_indx);
                Emit("        call    #_divide_\n");
                Emit1i("        mov     reg%d, parm1\n", reg_indx1);
                break;
            case AS_MOD:
                // TODO: Handle unsigned modulus
                Emit1i("        mov     parm1, reg%d\n", reg_indx1);
                Emit1i("        mov     parm2, reg%d\n", reg_indx);
                Emit("        call    #_modulus_\n");
                Emit1i("        mov     reg%d, parm1\n", reg_indx1);
                break;
            case AS_SHIFTL:
                Emit2("        shl     reg%d, reg%d\n", (char *)reg_indx1, (char *)reg_indx);
                break;
            case AS_SHIFTR:
                if (signflag)
                    Emit2("        sar     reg%d, reg%d\n", (char *)reg_indx1, (char *)reg_indx);
                else
                    Emit2("        shr     reg%d, reg%d\n", (char *)reg_indx1, (char *)reg_indx);
                break;
            case AS_AND:
                Emit2("        and     reg%d, reg%d\n", (char *)reg_indx, (char *)reg_indx1);
                reg_indx1 = reg_indx;
                break;
            case AS_OR:
                Emit2("        or      reg%d, reg%d\n", (char *)reg_indx, (char *)reg_indx1);
                reg_indx1 = reg_indx;
                break;
            case AS_XOR:
                Emit2("        xor     reg%d, reg%d\n", (char *)reg_indx, (char *)reg_indx1);
                reg_indx1 = reg_indx;
                break;
        }
    }
    else
    {
        reg_indx1 = reg_indx;
    }

    EmitWriteAccess(addrstr, typestr, parmchar, local, reg_indx1, absolute);

    if (retflag)
    {
        if (reg_indx1 != reg_indx)
            Emit2("        mov     reg%d, reg%d\n", (char *)reg_indx, (char *)reg_indx1);
        reg_indx++;
    }
}

void EmitCallPrep(int retflag)
{
    callstack[cs_index++] = reg_indx;
}

void EmitCall(int retflag, char *funcname)
{
    int i;
    int count = callstack[--cs_index];
#if 0
    for (i = 0; i < count; i++)
    {
        Emit("        sub     sp, #4\n");
        Emit1i("        wrlong  reg%d, sp\n", i);
    }
#endif
    if (count)
    {
#if 1
        Emit1i("        sub     sp, #%d\n", 4*count);
        if (count > 1)
            Emit1i("        setq    #%d\n", count-1);
        Emit("        wrlong  reg0, sp\n");
#endif
        for (i = count; i < reg_indx; i++)
        {
            Emit2("        mov     reg%d, reg%d\n", (char *)(i - count), (char *)i);
        }
    }
#ifdef USE_CALLD
    Emit1("        calld   adra, #%s\n", funcname);
#else
    Emit1("        call    #%s\n", funcname);
#endif
    EmitNops(0);
    reg_indx = count;
    if (retflag)
    {
        Emit1i("        mov     reg%d, reg0\n", count);
        reg_indx++;
    }
#if 0
    for (i = count - 1; i >= 0; i--)
    {
        Emit1i("        rdlong  reg%d, sp\n", i);
        Emit("        add     sp, #4\n");
    }
#else
    if (count)
    {
        if (count > 1)
            Emit1i("        setq    #%d\n", count-1);
        Emit("        rdlong  reg0, sp\n");
        Emit1i("        add     sp, #%d\n", 4*count);
    }
#endif
}

void EmitReturn(int stackspace)
{
    if (stackspace > 0)
    {
        if (stackspace < 512)
            Emit1i("        add     sp, #%d\n", stackspace);
        else
            Emit1i("        add     sp, ##%d\n", stackspace);
    }
#ifdef USE_CALLD
    Emit("        rdlong  adra, sp\n");
    Emit("        add     sp, #4\n");
#if 0
    Emit("        and     adra, ##$ffff\n");
    Emit("        calld   adra, adra\n");
#else
    Emit("        jmp     adra\n");
#endif
#else
    Emit("        ret\n");
#endif
    EmitNops(0);
}

void EmitReturnValue(int stackspace)
{
    if (reg_indx > 1)
        Emit1i("        mov     reg0, reg%d\n", reg_indx-1);
    if (stackspace > 0)
    {
        if (stackspace < 512)
            Emit1i("        add     sp, #%d\n", stackspace);
        else
            Emit1i("        add     sp, ##%d\n", stackspace);
    }
#ifdef USE_CALLD
    Emit("        rdlong  adra, sp\n");
    Emit("        add     sp, #4\n");
#if 0
    Emit("        and     adra, ##$ffff\n");
    Emit("        calld   adra, adra\n");
#else
    Emit("        jmp     adra\n");
#endif
#else
    Emit("        ret\n");
#endif
    EmitNops(0);
    reg_indx--;
}

void EmitEmbeddedString(char *str, int label1, int label2)
{
    char first = 1;
    char *ptr;
#if 0
    Emit1i("        jmp     #label%04d\n", label2);
    EmitNops(0);
    Emit1i("label%04d\n", label1);
#else
    Emit1i("        calld   adra, #label%04d\n", label2);
#endif
    Emit("        byte    ");
    if (*str == '"') str++;
    while (*str)
    {
        ptr = FindChars(str, "\"\\");
        if (ptr != str)
        {
            if (!first) Emit(", ");
            Emit("\"");
            while (ptr != str) Emit1i("%c", *str++);
            Emit("\"");
            first = 0;
        }
        if (*ptr == 0 || *ptr == '"')
        {
            if (!first) Emit(", ");
            Emit("0\n");
#if 0
            Emit("        orgh    ($ + 3) & (-4)\n");
#else
            Emit("        alignl\n");
#endif
            break;
        }
        str++;
        if (*str)
        {
            if (!first) Emit(", ");
            if (*str == 'n')
                Emit("10");
            else if (*str == 'r')
                Emit("13");
            else
                Emit1i("\"\\%c\"", *str);
            str++;
            first = 0;
        }
    }
    Emit1i("label%04d\n", label2);
#if 0
    Emit2("        mov     reg%d, ##label%04d\n", (char *)(reg_indx++), (char *)label1);
    if (picflag)
        Emit1("        add     reg%d, ptrb\n", (char *)(reg_indx-1));
#else
    Emit1("        mov     reg%d, adra\n", (char *)(reg_indx++));
#endif
}

void EmitLabelString(char *str)
{
    Emit1("%s\n", str);
}

void EmitParmStart(void)
{
}

void EmitParmEnd(void)
{
}

void EmitLocalVar(char *varname)
{
}

void EmitNumParmsCode(int numparms)
{
    reg_indx += numparms;
}

void EmitAlignment(int size)
{
    if (size == 4)
        Emit("        alignl\n");
    else if (size == 2)
        Emit("        alignw\n");
}

void EmitFunctionStart(char *funcname)
{
    if (globalflag)
        Emit1("%-7s global\n", funcname);
    else
        Emit1("%-7s\n", funcname);
    //EmitLabelString(funcname);
#ifdef USE_CALLD
    Emit("        sub     sp, #4\n");
    Emit("        wrlong  adra, sp\n");
#endif
    //Emit("        sub     sp, #100\n");
    reg_indx = 0;
}

void EmitAllocateStackSpace(int stackspace)
{
    if (stackspace > 0)
    {
        if (stackspace < 512)
            Emit1i("        sub     sp, #%d\n", stackspace);
        else
            Emit1i("        sub     sp, ##%d\n", stackspace);
    }
}

void EmitIntrinsic(char *funcname)
{
    Emit2("        %-7s reg%d\n", funcname, (char *)(--reg_indx));
}

void EmitQuotedString(char *str)
{
    int len;
    if (str[0] == '"') str++;
    len = strlen(str) - 1;
    if (len >= 0 && str[len] == '"') str[len] = 0;
    Emit1("%s\n", str);
}

/*
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
*/
